#ifndef lint
static const char rcsid[] = "$Header: /mona/tristan/mpeg/dec/RCS/multi_player.c,v 1.1 1998/08/29 06:58:26 tristan Exp $";
static const char copyright[] = "Copyright (c) 1998, MpegTV LLC";
#endif


/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  Copyright-holder disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 */

#include "mtvp/mtvp.h"

/* challenge.h not available in public distribution */
#ifdef HANDLE_CHALLENGE
#	include "challenge.h"
#endif

/*--------------------------------------------------------------------------+
| Configuration flags														|
+--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------+
| Macros																	|
+--------------------------------------------------------------------------*/
/* for the msg.h package */
#define MSG_PTR (&Msg)

/* maximum number of streams that we want to play at the same time! */
#define MAX_PLAYERS	10

/*--------------------------------------------------------------------------+
| Data types																|
+--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------+
| Globals																	|
+--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------+
| Statics																	|
+--------------------------------------------------------------------------*/

/* to communicate with the players */
static PMP_STRUCT	Player[MAX_PLAYERS];

static PMP_STRUCT	*Player_with_audio;

/* for the msg.h package */
static MSG_STRUCT Msg;


/*--------------------------------------------------------------------------+
| read_and_handle_message_from_player										|
+--------------------------------------------------------------------------*/
static int
read_and_handle_message_from_player(PMP_STRUCT *player)
{
	int		fd = player->receive_channel.msg_pipe[PMP_FD_READ];
	PMP_HEADER	msg;		/* to hold message header from Player */ 
	PMP_DATA_A2G	data;	/* to hold message data from Player */
	
	/*
	 * This  callback is called when there is something available for
	 * reading on Player.receive_channel.msg_pipe[PMP_FD_READ]
	 * (also equal to fd in this routine).
	 *
	 * This routine should read one message from the Player
	 * and take the appropriate action.
	 *
	 * It is mostly a place holder in this example: This sample
	 * application ignores most messages.  It just checks the
	 * API version number and quits if the message A2G_COM_QUIT
	 * is received (which should not happen, by the way!).
	 */
	 
	/* read message from player */
	if (a2g_read_msg(player, &msg, &data) == -1) {
		return -1;
	}

	/* trace message if (Msg_select & MSG_INFO) */
	{
		char buf[PMP_PRINT_MSG_BUF_SIZE];
		MESSAGE(("read_and_handle_message_from_player: received msg: %s", pmp_print_msg(&msg,  (const char *)&data, buf)));
	}
	
	switch ((A2G_COM)msg.type) {
		case A2G_COM_QUIT:
			/*
			 * Player sends A2G_COM_QUIT after receiving COM_QUIT and before exiting.
			 */

			/* g2a_cleanup() should be called to close pipes to the (now defunct) player */
			g2a_cleanup(player);
			
			break;
		case A2G_COM_X_EVENT:			
			/*
			 * Player sends A2G_COM_X_EVENT to forward the X Events that we have
			 * requested (see COM_X_EVENT_MASK).
			 */
			switch (data.data_x_event.type) {
				
				case ButtonRelease:
					/*
					 * This is a click in the video window of this player.
					 * Tell all other players to mute (since we don't know which
					 * one plays the audio), then tell this player to un-mute.
					 */
					{
						u_int i;
						for (i = 0; i < MAX_PLAYERS; i++) {
							int fd = Player[i].receive_channel.msg_pipe[PMP_FD_READ];
				
							if (&Player[i] != player && fd != -1) {
								g2a_send_int(&Player[i], COM_AUDIO_MUTE, 0, (int)True);
							}
						}
					}
					g2a_send_int(player, COM_AUDIO_MUTE, 0, (int)False);
					
					/* note what player is supposed to be playing the audio */
					Player_with_audio = player;
					break;
				default:
					/* ignore other X events */
					break;
			}
			break;
		case A2G_COM_AUDIO_MUTE:
			/*
			 * Player sends A2G_COM_AUDIO_MUTE when an un-muted player
			 * is not able to open the audio device (e.g. device is busy).
			 */
			if (player == Player_with_audio) {
				/* if that comes from the player that was selected for audio, try again */
				g2a_send_int(player, COM_AUDIO_MUTE, 0, (int)False);
			}
			break;
		case A2G_COM_CHALLENGE:
			/*
			 * This is a reserved message of the API.
			 * A2G_COM_CHALLENGE is a private authentification request
			 * which is sent only if the player is started with option -k
			 * or cannot find a valid license key.
			 * The player does not need a valid license key if the front-end
			 * can be authentified. Authentification is done by responding
			 * properly to challenge messages.
			 *
			 * If the player finds a valid license key, it does not
			 * send challenge messages.
			 *
			 * Front-ends who receive challenge messages but are not capable of
			 * dealing with them should advise the user that the player has no
			 * license and will not be functionnal.
			 */

#ifdef HANDLE_CHALLENGE
			if (handle_challenge(player, &msg, &data) != 0) {
				FATAL_ERROR(("cannot handle challenge"));
			}
#else
			FATAL(("This front-end is not authentified by mtvp.\nYou need a development license key (in file mtvp.key).\nSee www.mpegtv.com/sdk.html for details."));

#endif
			break;
		default:
			/* ignore all other messages */
			break;
	}

	return 0;
}


/*--------------------------------------------------------------------------+
| clean_exit																|
+--------------------------------------------------------------------------*/
static void
clean_exit(int n)
{	
	MESSAGE(("(pid=%u): exiting, status = %d", (u_int)getpid(), n));

	/* terminate (real exit this time) */
	exit(n);
}


/*--------------------------------------------------------------------------+
| start_player																|
+--------------------------------------------------------------------------*/
static void
start_player(PMP_STRUCT *player, const char **argv)
{
	static const char *mtvp_argv[] = {
		"../../bin/mtvp",	/* name of the player (if no / in pathname, will search in PATH) */
		/* "-a0", */	/* possible options */
		/* "-w", */	/* possible options */
#ifdef HANDLE_CHALLENGE
		"-k",	/* use challenge authentification only, do not look for key file */
#endif
		NULL,	/* must be NULL-terminated */
	};

	/*
	 * First parameter of g2a_init() is a ptr to a NULL-terminated array
	 * of strings.
	 *
	 * The first string is the name of the Player ("mtvp").
	 * (when using libmtvpnew.so, this string is ignored!)
	 * The following strings are possible parameters.
	 *
	 * If the first string is a full pathname (starting with "/")
	 * g2a_init() will run this program as "the Player".
	 *
	 * If the first string is not a full pathname (starting with "/")
	 * g2a_init() will look for the Player in the following places:
	 * - in the directory where argv[0] is located (if argv[0] contains a '/')
	 * - in the current directory (".")
	 * - in the PATH
	 *
	 * Typically first string is "mtvp", if the player was installed
	 * under this name.
	 *
	 * Second parameter of g2a_init() is a string that should be
	 * the argv[0] from the calling program.
	 */

	if (g2a_init(player, mtvp_argv, argv[0]) == -1) {
		if (errno == 0) {
			FATAL_ERROR(("cannot start player (%s) - probably mtvp API version mismatch", mtvp_argv[0]));
		} else if (errno == EPIPE) {
			FATAL_ERROR(("cannot start player (%s) - check that player is in your PATH!", mtvp_argv[0]));
		} else {
			FATAL_ERROR(("cannot start player (%s)", mtvp_argv[0]));
		}
	}
}

/*--------------------------------------------------------------------------+
| main																		|
+--------------------------------------------------------------------------*/
/*ARGSUSED*/
int
main(
	int argc,
	const char **argv,
	const char **envp
)
{
	const char	**p, **q;
	int		n;

	/*
	 * Set Msg.select == -1 to print all the warnings and messages.
	 * See msg.h for the MESSAGE() and WARNING() macros definitions.
	 */
	 
	/* Msg.select = -1; */
	
	MESSAGE(("started, pid = %u", (u_int)getpid()));
	
	if (argc <= 1) {
		fprintf(stderr, "usage: multi_player file1.mpg [file2.mpg...]\n");
		clean_exit(EXIT_FAILURE);		
	}

	/*
	 * Start a player for each of the files specified, and tell the player
	 * to play the file in loop.
	 */
	
	{
		u_int	i;
		
		for (i = 0; i < MAX_PLAYERS; i++) {

			/*
			 * If player is not started, set fd to -1.  Then we know,
			 * later, in the select, which fd we should listen to.
			 */
			Player[i].receive_channel.msg_pipe[PMP_FD_READ] = -1;

			if (i <= (u_int)(argc - 2)) {
			
				/* start a player */
				start_player(&Player[i], argv);
				
				/* tell player to loop when end of file will be reached */
				g2a_send_int(&Player[i], COM_LOOP, 0, (int)True);

				/*
				 * Tell player to automatically fillup the window.
				 * User can scale the video by resizing the window interactively.
				 */
				g2a_send_int(&Player[i], COM_VIDEO_LAYOUT, 0, VIDEO_LAYOUT_FILL_WINDOW);
				
				/*
				 * Tell player to pass us buttons events.  When a click occurs
				 * in a video window, we will tell the corresponding player to
				 * play the audio.
				 */
				g2a_send_int(&Player[i], COM_X_EVENT_MASK, 0, (int)(ButtonPressMask|ButtonReleaseMask));

				/* tell player to open file */
				g2a_send_string(&Player[i], COM_OPEN_FILE, 0, argv[i + 1]);
			}
		}
	}
			
	MESSAGE(("entering event loop"));

	/* enter event loop */
	while (True) {
		u_int	i;
		fd_set	rfds;
		int		ret;
		int		maxfd = -1;

		/* initialize rfds structures with players fd */
		FD_ZERO(&rfds);
		for (i = 0; i < MAX_PLAYERS; i++) {
			int fd = Player[i].receive_channel.msg_pipe[PMP_FD_READ];
			
			if (fd != -1) {
				FD_SET(fd, &rfds);
				/* remember the largest of the fd's */
				if (fd > maxfd) {
					maxfd = fd;
				}
			}
		}
		if (maxfd == -1) {
			/* no players ? exit loop! */
			break;
		}

		/* block until data is available on one of the players fd's */
		if ((ret = select(maxfd + 1, &rfds, NULL, NULL, NULL /* &tv */)) == -1) {
			if (errno == EINTR) {	/* ignore this */
				continue;
			}
			FATAL_ERROR(("select failed"));
		}
		/* read the player's message on each fd where a message is available */
		for (i = 0; i < MAX_PLAYERS; i++) {
			int fd = Player[i].receive_channel.msg_pipe[PMP_FD_READ];

			if (fd != -1 && FD_ISSET(fd, &rfds)) {
				if (read_and_handle_message_from_player(&Player[i]) != 0) {
					/* player probably dead, stop listening to this fd! */
					Player[i].receive_channel.msg_pipe[PMP_FD_READ] = -1;
				}
			}
		}
	}

	/* terminated successfully */
	clean_exit(EXIT_SUCCESS);

	/* so that lint does not complain! */
	return 0;
}


/* eof */
